@@ -35,44 +35,10 @@
 #include <X11/Xatom.h>
 
 #include "private.h"
+#include "gnome-rr-private.h"
 
 #define DISPLAY(o) ((o)->info->screen->xdisplay)
 
-typedef struct ScreenInfo ScreenInfo;
-
-struct ScreenInfo
-{
-    int			min_width;
-    int			max_width;
-    int			min_height;
-    int			max_height;
-
-    XRRScreenResources *resources;
-    
-    GnomeRROutput **	outputs;
-    GnomeRRCrtc **	crtcs;
-    GnomeRRMode **	modes;
-    
-    GnomeRRScreen *	screen;
-
-    GnomeRRMode **	clone_modes;
-};
-
-struct GnomeRRScreen
-{
-    GdkScreen *			gdk_screen;
-    GdkWindow *			gdk_root;
-    Display *			xdisplay;
-    Screen *			xscreen;
-    Window			xroot;
-    ScreenInfo *		info;
-    
-    int				randr_event_base;
-    
-    GnomeRRScreenChanged	callback;
-    gpointer			data;
-};
-
 struct GnomeRROutput
 {
     ScreenInfo *	info;
@@ -325,131 +291,161 @@ gather_clone_modes (ScreenInfo *info)
     info->clone_modes = (GnomeRRMode **)g_ptr_array_free (result, FALSE);
 }
 
+static gboolean
+fill_screen_info_from_resources (ScreenInfo *info,
+				 XRRScreenResources *resources,
+				 GError **error)
+{
+    int i;
+    GPtrArray *a;
+    GnomeRRCrtc **crtc;
+    GnomeRROutput **output;
+
+    info->resources = resources;
+
+    /* We create all the structures before initializing them, so
+     * that they can refer to each other.
+     */
+    a = g_ptr_array_new ();
+    for (i = 0; i < resources->ncrtc; ++i)
+    {
+	GnomeRRCrtc *crtc = crtc_new (info, resources->crtcs[i]);
+
+	g_ptr_array_add (a, crtc);
+    }
+    g_ptr_array_add (a, NULL);
+    info->crtcs = (GnomeRRCrtc **)g_ptr_array_free (a, FALSE);
+
+    a = g_ptr_array_new ();
+    for (i = 0; i < resources->noutput; ++i)
+    {
+	GnomeRROutput *output = output_new (info, resources->outputs[i]);
+
+	g_ptr_array_add (a, output);
+    }
+    g_ptr_array_add (a, NULL);
+    info->outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
+
+    a = g_ptr_array_new ();
+    for (i = 0;  i < resources->nmode; ++i)
+    {
+	GnomeRRMode *mode = mode_new (info, resources->modes[i].id);
+
+	g_ptr_array_add (a, mode);
+    }
+    g_ptr_array_add (a, NULL);
+    info->modes = (GnomeRRMode **)g_ptr_array_free (a, FALSE);
+
+    /* Initialize */
+    for (crtc = info->crtcs; *crtc; ++crtc)
+    {
+	if (!crtc_initialize (*crtc, resources, error))
+	    return FALSE;
+    }
+
+    for (output = info->outputs; *output; ++output)
+    {
+	if (!output_initialize (*output, resources, error))
+	    return FALSE;
+    }
+
+    for (i = 0; i < resources->nmode; ++i)
+    {
+	GnomeRRMode *mode = mode_by_id (info, resources->modes[i].id);
+
+	mode_initialize (mode, &(resources->modes[i]));
+    }
+
+    gather_clone_modes (info);
+
+    return TRUE;
+}
+
 static gboolean
 fill_out_screen_info (Display *xdisplay,
 		      Window xroot,
 		      ScreenInfo *info,
+		      gboolean needs_reprobe,
 		      GError **error)
 {
     XRRScreenResources *resources;
     
     g_assert (xdisplay != NULL);
     g_assert (info != NULL);
-    
-    gdk_error_trap_push ();
-    
-    if (!XRRGetScreenSizeRange (xdisplay, xroot,
-                                &(info->min_width),
-                                &(info->min_height),
-                                &(info->max_width),
-                                &(info->max_height))) {
-        /* XRR caught an error */
-	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
-		     _("could not get the range of screen sizes"));
-        return FALSE;
-    }
-    
-    gdk_flush ();
-    if (gdk_error_trap_pop ())
+
+    /* First update the screen resources */
+
+    if (needs_reprobe)
+        resources = XRRGetScreenResources (xdisplay, xroot);
+    else
     {
-        /* Unhandled X Error was generated */
-	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_UNKNOWN,
-		     _("unhandled X error while getting the range of screen sizes"));
-        return FALSE;
-    }
-    
-#if 0
-    g_print ("ranges: %d - %d; %d - %d\n",
-	     screen->min_width, screen->max_width,
-	     screen->min_height, screen->max_height);
+	/* XRRGetScreenResourcesCurrent is less expensive than
+	 * XRRGetScreenResources, however it is available only
+	 * in RandR 1.3 or higher
+	 */
+#if (RANDR_MAJOR > 1 || (RANDR_MAJOR == 1 && RANDR_MINOR >= 3))
+        /* Runtime check for RandR 1.3 or higher */
+        if (info->screen->rr_major_version == 1 && info->screen->rr_minor_version >= 3)
+            resources = XRRGetScreenResourcesCurrent (xdisplay, xroot);
+        else
+            resources = XRRGetScreenResources (xdisplay, xroot);
+#else
+        resources = XRRGetScreenResources (xdisplay, xroot);
 #endif
-    
-    resources = XRRGetScreenResources (xdisplay, xroot);
-    
+    }
+
     if (resources)
     {
-	int i;
-	GPtrArray *a;
-	GnomeRRCrtc **crtc;
-	GnomeRROutput **output;
-	
-#if 0
-	g_print ("Resource Timestamp: %u\n", (guint32)resources->timestamp);
-	g_print ("Resource Configuration Timestamp: %u\n", (guint32)resources->configTimestamp);
-#endif
-	
-	info->resources = resources;
-	
-	/* We create all the structures before initializing them, so
-	 * that they can refer to each other.
-	 */
-	a = g_ptr_array_new ();
-	for (i = 0; i < resources->ncrtc; ++i)
-	{
-	    GnomeRRCrtc *crtc = crtc_new (info, resources->crtcs[i]);
-	    
-	    g_ptr_array_add (a, crtc);
-	}
-	g_ptr_array_add (a, NULL);
-	info->crtcs = (GnomeRRCrtc **)g_ptr_array_free (a, FALSE);
-	
-	a = g_ptr_array_new ();
-	for (i = 0; i < resources->noutput; ++i)
-	{
-	    GnomeRROutput *output = output_new (info, resources->outputs[i]);
-	    
-	    g_ptr_array_add (a, output);
-	}
-	g_ptr_array_add (a, NULL);
-	info->outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
-	
-	a = g_ptr_array_new ();
-	for (i = 0;  i < resources->nmode; ++i)
-	{
-	    GnomeRRMode *mode = mode_new (info, resources->modes[i].id);
-	    
-	    g_ptr_array_add (a, mode);
-	}
-	g_ptr_array_add (a, NULL);
-	info->modes = (GnomeRRMode **)g_ptr_array_free (a, FALSE);
-	
-	/* Initialize */
-	for (crtc = info->crtcs; *crtc; ++crtc)
-	{
-	    if (!crtc_initialize (*crtc, resources, error))
-		return FALSE;
-	}
-	
-	for (output = info->outputs; *output; ++output)
-	{
-	    if (!output_initialize (*output, resources, error))
-		return FALSE;
-	}
-	
-	for (i = 0; i < resources->nmode; ++i)
-	{
-	    GnomeRRMode *mode = mode_by_id (info, resources->modes[i].id);
-	    
-	    mode_initialize (mode, &(resources->modes[i]));
-	}
-
-	gather_clone_modes (info);
-	
-	return TRUE;
+	if (!fill_screen_info_from_resources (info, resources, error))
+	    return FALSE;
     }
     else
     {
-#if 0
-	g_print ("Couldn't get screen resources\n");
-#endif
 	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
 		     _("could not get the screen resources (CRTCs, outputs, modes)"));
 	return FALSE;
     }
+
+    /* Then update the screen size range.  We do this after XRRGetScreenResources() so that
+     * the X server will already have an updated view of the outputs.
+     */
+
+    if (needs_reprobe) {
+	gboolean success;
+
+        gdk_error_trap_push ();
+	success = XRRGetScreenSizeRange (xdisplay, xroot,
+					 &(info->min_width),
+					 &(info->min_height),
+					 &(info->max_width),
+					 &(info->max_height));
+	gdk_flush ();
+	if (gdk_error_trap_pop ()) {
+	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_UNKNOWN,
+			 _("unhandled X error while getting the range of screen sizes"));
+	    return FALSE;
+	}
+
+	if (!success) {
+	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+			 _("could not get the range of screen sizes"));
+            return FALSE;
+        }
+    }
+    else
+    {
+        gnome_rr_screen_get_ranges (info->screen, 
+					 &(info->min_width),
+					 &(info->max_width),
+					 &(info->min_height),
+					 &(info->max_height));
+    }
+
+    return TRUE;
 }
 
 static ScreenInfo *
-screen_info_new (GnomeRRScreen *screen, GError **error)
+screen_info_new (GnomeRRScreen *screen, gboolean needs_reprobe, GError **error)
 {
     ScreenInfo *info = g_new0 (ScreenInfo, 1);
     
@@ -460,7 +456,7 @@ screen_info_new (GnomeRRScreen *screen, GError **error)
     info->modes = NULL;
     info->screen = screen;
     
-    if (fill_out_screen_info (screen->xdisplay, screen->xroot, info, error))
+    if (fill_out_screen_info (screen->xdisplay, screen->xroot, info, needs_reprobe, error))
     {
 	return info;
     }
@@ -472,14 +468,14 @@ screen_info_new (GnomeRRScreen *screen, GError **error)
 }
 
 static gboolean
-screen_update (GnomeRRScreen *screen, gboolean force_callback, GError **error)
+screen_update (GnomeRRScreen *screen, gboolean force_callback, gboolean needs_reprobe, GError **error)
 {
     ScreenInfo *info;
     gboolean changed = FALSE;
     
     g_assert (screen != NULL);
 
-    info = screen_info_new (screen, error);
+    info = screen_info_new (screen, needs_reprobe, error);
     if (!info)
 	    return FALSE;
 
@@ -503,23 +499,88 @@ screen_on_event (GdkXEvent *xevent,
 {
     GnomeRRScreen *screen = data;
     XEvent *e = xevent;
-    
-    if (e && e->type - screen->randr_event_base == RRNotify)
+    int event_num;
+
+    if (!e)
+	return GDK_FILTER_CONTINUE;
+
+    event_num = e->type - screen->randr_event_base;
+
+    if (event_num == RRScreenChangeNotify) {
+	/* We don't reprobe the hardware; we just fetch the X server's latest
+	 * state.  The server already knows the new state of the outputs; that's
+	 * why it sent us an event!
+	 */
+        screen_update (screen, TRUE, FALSE, NULL); /* NULL-GError */
+#if 0
+	/* Enable this code to get a dialog showing the RANDR timestamps, for debugging purposes */
+	{
+	    GtkWidget *dialog;
+	    XRRScreenChangeNotifyEvent *rr_event;
+	    static int dialog_num;
+
+	    rr_event = (XRRScreenChangeNotifyEvent *) e;
+
+	    dialog = gtk_message_dialog_new (NULL,
+					     0,
+					     GTK_MESSAGE_INFO,
+					     GTK_BUTTONS_CLOSE,
+					     "RRScreenChangeNotify timestamps (%d):\n"
+					     "event change: %u\n"
+					     "event config: %u\n"
+					     "event serial: %lu\n"
+					     "----------------------"
+					     "screen change: %u\n"
+					     "screen config: %u\n",
+					     dialog_num++,
+					     (guint32) rr_event->timestamp,
+					     (guint32) rr_event->config_timestamp,
+					     rr_event->serial,
+					     (guint32) screen->info->resources->timestamp,
+					     (guint32) screen->info->resources->configTimestamp);
+	    g_signal_connect (dialog, "response",
+			      G_CALLBACK (gtk_widget_destroy), NULL);
+	    gtk_widget_show (dialog);
+	}
+#endif
+    }
+#if 0
+    /* WHY THIS CODE IS DISABLED:
+     *
+     * Note that in gnome_rr_screen_new(), we only select for
+     * RRScreenChangeNotifyMask.  We used to select for other values in
+     * RR*NotifyMask, but we weren't really doing anything useful with those
+     * events.  We only care about "the screens changed in some way or another"
+     * for now.
+     *
+     * If we ever run into a situtation that could benefit from processing more
+     * detailed events, we can enable this code again.
+     *
+     * Note that the X server sends RRScreenChangeNotify in conjunction with the
+     * more detailed events from RANDR 1.2 - see xserver/randr/randr.c:TellChanged().
+     */
+    else if (event_num == RRNotify)
     {
+	/* Other RandR events */
+
 	XRRNotifyEvent *event = (XRRNotifyEvent *)e;
-	
+
+	/* Here we can distinguish between RRNotify events supported
+	 * since RandR 1.2 such as RRNotify_OutputProperty.  For now, we
+	 * don't have anything special to do for particular subevent types, so
+	 * we leave this as an empty switch().
+	 */
 	switch (event->subtype)
 	{
 	default:
 	    break;
 	}
-	
-	/* FIXME: we may need to be more discriminating in
-	 * what causes 'changed' events
-	 */
-	screen_update (screen, TRUE, NULL); /* NULL-GError */
+
+	/* No need to reprobe hardware here */
+	screen_update (screen, TRUE, FALSE, NULL); /* NULL-GError */
     }
-    
+#endif
+
     /* Pass the event on to GTK+ */
     return GDK_FILTER_CONTINUE;
 }
@@ -555,26 +616,28 @@ gnome_rr_screen_new (GdkScreen *gdk_screen,
 	screen->data = data;
 	
 	screen->randr_event_base = event_base;
-	
-	screen->info = screen_info_new (screen, error);
+
+	XRRQueryVersion (dpy, &screen->rr_major_version, &screen->rr_minor_version);
+
+	screen->info = screen_info_new (screen, TRUE, error);
 	
 	if (!screen->info) {
 	    g_free (screen);
 	    return NULL;
 	}
-	
-	XRRSelectInput (screen->xdisplay,
-			screen->xroot,
-			RRScreenChangeNotifyMask	|
-			RRCrtcChangeNotifyMask		|
-			RROutputPropertyNotifyMask);
-	
-	gdk_x11_register_standard_event_type (
-	    gdk_screen_get_display (gdk_screen),
-	    event_base,
-	    RRNotify + 1);
-	
-	gdk_window_add_filter (screen->gdk_root, screen_on_event, screen);
+
+	if (screen->callback) {
+	    XRRSelectInput (screen->xdisplay,
+			    screen->xroot,
+			    RRScreenChangeNotifyMask);
+
+	    gdk_x11_register_standard_event_type (gdk_screen_get_display (gdk_screen),
+						  event_base,
+						  RRNotify + 1);
+
+	    gdk_window_add_filter (screen->gdk_root, screen_on_event, screen);
+	}
+
 	return screen;
     }
     else
@@ -634,6 +697,33 @@ gnome_rr_screen_get_ranges (GnomeRRScreen *screen,
 	*max_height = screen->info->max_height;
 }
 
+/**
+ * gnome_rr_screen_get_timestamps
+ * @screen: a #GnomeRRScreen
+ * @change_timestamp_ret: Location in which to store the timestamp at which the RANDR configuration was last changed
+ * @config_timestamp_ret: Location in which to store the timestamp at which the RANDR configuration was last obtained
+ *
+ * Queries the two timestamps that the X RANDR extension maintains.  The X
+ * server will prevent change requests for stale configurations, those whose
+ * timestamp is not equal to that of the latest request for configuration.  The
+ * X server will also prevent change requests that have an older timestamp to
+ * the latest change request.
+ */
+void
+gnome_rr_screen_get_timestamps (GnomeRRScreen *screen,
+				guint32       *change_timestamp_ret,
+				guint32       *config_timestamp_ret)
+{
+    g_return_if_fail (screen != NULL);
+
+    if (change_timestamp_ret)
+	*change_timestamp_ret = screen->info->resources->timestamp;
+
+    if (config_timestamp_ret)
+	*config_timestamp_ret = screen->info->resources->configTimestamp;
+}
+
+
 /**
  * gnome_rr_screen_refresh
  * @screen: a #GnomeRRScreen
@@ -652,7 +742,7 @@ gnome_rr_screen_refresh (GnomeRRScreen *screen,
 			 GError       **error)
 {
     g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
-    return screen_update (screen, FALSE, error);
+    return screen_update (screen, FALSE, TRUE, error);
 }
 
 GnomeRRMode **
@@ -776,13 +866,21 @@ get_property (Display *dpy,
 static guint8 *
 read_edid_data (GnomeRROutput *output)
 {
-    Atom edid_atom = XInternAtom (DISPLAY (output), "EDID_DATA", FALSE);
+    Atom edid_atom;
     guint8 *result;
     int len;
-    
+
+    edid_atom = XInternAtom (DISPLAY (output), "EDID", FALSE);
     result = get_property (DISPLAY (output),
 			   output->id, edid_atom, &len);
-    
+
+    if (!result)
+    {
+	edid_atom = XInternAtom (DISPLAY (output), "EDID_DATA", FALSE);
+	result = get_property (DISPLAY (output),
+			       output->id, edid_atom, &len);
+    }
+
     if (result)
     {
 	if (len % 128 == 0)
@@ -809,6 +907,7 @@ output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **erro
     if (!info || !output->info)
     {
 	/* FIXME: see the comment in crtc_initialize() */
+	/* Translators: here, an "output" is a video output */
 	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
 		     _("could not get information about output %d"),
 		     (int) output->id);
@@ -1086,6 +1185,20 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
 			  GnomeRROutput   **outputs,
 			  int               n_outputs,
 			  GError          **error)
+{
+    return gnome_rr_crtc_set_config_with_time (crtc, GDK_CURRENT_TIME, x, y, mode, rotation, outputs, n_outputs, error);
+}
+
+gboolean
+gnome_rr_crtc_set_config_with_time (GnomeRRCrtc      *crtc,
+				    guint32           timestamp,
+				    int               x,
+				    int               y,
+				    GnomeRRMode      *mode,
+				    GnomeRRRotation   rotation,
+				    GnomeRROutput   **outputs,
+				    int               n_outputs,
+				    GError          **error)
 {
     ScreenInfo *info;
     GArray *output_ids;
@@ -1107,7 +1220,7 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
 	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR,
 			 /* Translators: the "position", "size", and "maximum"
 			  * words here are not keywords; please translate them
-			  * as usual.  */
+			  * as usual.  A CRTC is a CRT Controller (this is X terminology) */
 			 _("requested position/size for CRTC %d is outside the allowed limit: "
 			   "position=(%d, %d), size=(%d, %d), maximum=(%d, %d)"),
 			 (int) crtc->id,
@@ -1127,7 +1240,7 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
     }
     
     status = XRRSetCrtcConfig (DISPLAY (crtc), info->resources, crtc->id,
-			       CurrentTime, 
+			       timestamp, 
 			       x, y,
 			       mode ? mode->id : None,
 			       xrotation_from_rotation (rotation),
@@ -1140,6 +1253,9 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
 	result = TRUE;
     else {
 	result = FALSE;
+	/* Translators: CRTC is a CRT Controller (this is X terminology).
+	 * It is *very* unlikely that you'll ever get this error, so it is
+	 * only listed for completeness. */
 	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
 		     _("could not set the configuration for CRTC %d"),
 		     (int) crtc->id);
@@ -1249,6 +1365,9 @@ crtc_initialize (GnomeRRCrtc        *crtc,
 	/* FIXME: We need to reaquire the screen resources */
 	/* FIXME: can we actually catch BadRRCrtc, and does it make sense to emit that? */
 
+	/* Translators: CRTC is a CRT Controller (this is X terminology).
+	 * It is *very* unlikely that you'll ever get this error, so it is
+	 * only listed for completeness. */
 	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
 		     _("could not get information about CRTC %d"),
 		     (int) crtc->id);
